from pylab import *
from numpy import *
import pandas as pd

import time

import cPickle as pickle
import tables as tbl
import h5py
import sqlite3 as sql
import persistent, transaction, ZODB, BTrees.IOBTree

from sqlalchemy import Table, Column, Integer
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


SqlBase = declarative_base()

class SqlRow(SqlBase):
  __tablename__ = 'data'
  c1 = Column(Integer, primary_key = True)  # Primary key is 'fake', but required by ORM
  c2 = Column(Integer)
  c3 = Column(Integer)
  c4 = Column(Integer)
  c5 = Column(Integer)


def mk_sqlite(filename, nrow = 10000):
  db = sql.connect(filename)
  c = db.cursor()
  c.execute("CREATE TABLE data (c1 REAL, c2 REAL, c3 REAL, c4 REAL, c5 REAL)")
  for i in xrange(nrow):
    c.execute("INSERT INTO data VALUES (?, ?, ?, ?, ?)", random.uniform(0.0, 100.0, 5))
  db.commit()
  db.close()


class TableRow(tbl.IsDescription):
  c1 = tbl.FloatCol()
  c2 = tbl.FloatCol()
  c3 = tbl.FloatCol()
  c4 = tbl.FloatCol()
  c5 = tbl.FloatCol()

def mk_table(filename, nrow = 10000):
  h5f = tbl.open_file(filename, mode='w', title='Benchmark')
  t = h5f.create_table('/', 'data', TableRow, 'Random data')
  row = t.row
  for i in xrange(nrow):
    for cname in ('c1', 'c2', 'c3', 'c4', 'c5'):
      row[cname] = random.uniform(0.0, 100.0)
    row.append()
  h5f.close()


def mk_csv(filename, nrow = 10000):
  with open(filename, 'w') as fp:
    for i in xrange(nrow):
      print >>fp, ','.join(str(ci) for ci in random.uniform(0.0, 100.0, 5))


def mk_pickle(filename, nrow = 10000):
  with open(filename, 'wb') as fp:
    pickle.dump(nrow, fp, pickle.HIGHEST_PROTOCOL)  # Write dataset length
    for i in xrange(nrow):
      row = tuple(random.uniform(0.0, 100.0, 5))
      pickle.dump(row, fp, pickle.HIGHEST_PROTOCOL)


class Row(object):
  def __init__(self, c1, c2, c3, c4, c5):
    self.c1 = c1
    self.c2 = c2
    self.c3 = c3
    self.c4 = c4
    self.c5 = c5

def mk_pickle_obj(filename, nrow = 10000):
  with open(filename, 'wb') as fp:
    pickle.dump(nrow, fp, pickle.HIGHEST_PROTOCOL)  # Write dataset length
    for i in xrange(nrow):
      row = Row(*tuple(random.uniform(0.0, 100.0, 5)))
      pickle.dump(row, fp, pickle.HIGHEST_PROTOCOL)


# In separate file, so it works across bench and bench_pypy
from zrow import ZRow

def mk_zodb(filename, nrow = 10000):
  conn = ZODB.connection(filename)
  #data = BTrees.OOBTree.BTree()  # less efficient, more general
  data = BTrees.IOBTree.BTree()
  conn.root.data = data
  for i in xrange(nrow):
    data[i] = ZRow(*tuple(random.uniform(0.0, 100.0, 5)))
  transaction.commit()
  conn.close()


def scan_sqlite(filename, niter = 100):
  start = time.time()
  db = sql.connect(filename)
  for it in xrange(niter):
    c = db.cursor()
    c.execute("SELECT * FROM data")
    row = c.fetchone()
    while row is not None:
      #print row
      row = c.fetchone()
  db.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)    
  return t/niter

def scan_sqlite_obj(filename, niter = 100):
  start = time.time()
  db = sql.connect(filename)
  for it in xrange(niter):
    c = db.cursor()
    c.execute("SELECT * FROM data")
    row = c.fetchone()
    while row is not None:
      row_obj = Row(*row)
      #print row_obj
      row = c.fetchone()
  db.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)    
  return t/niter


def scan_sqlalchemy_orm(filename, niter = 100):
  start = time.time()
  engine = create_engine('sqlite:///%s' % filename)
  Session = sessionmaker(bind=engine)
  for it in xrange(niter):
    session = Session()
    for row in session.query(SqlRow):
      #print row
      pass
    session.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)    
  return t/niter

def scan_sqlalchemy_core(filename, niter = 100):
  start = time.time()
  engine = create_engine('sqlite:///%s' % filename)
  for it in xrange(niter):
    conn = engine.connect()
    for row in conn.execute(SqlRow.__table__.select()):
      #print row
      pass
    conn.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter


def scan_table(filename, niter = 10):
  start = time.time()
  h5f = tbl.open_file(filename, mode='r')
  for it in xrange(niter):
    nrows = len(h5f.root.data)
    data = h5f.root.data
    #for i in xrange(nrows):
    #  row = data[i]
    #  #print row
    for row in data:
      rowdata = row.fetch_all_fields()
      #print rowdata
  h5f.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter


def scan_h5(filename, niter = 10):
  start = time.time()
  for it in xrange(niter):
    h5f = h5py.File(filename)
    nrows = h5f['/data'].len()
    data = h5f['/data']
    for row in data:
      #print row
      pass
    h5f.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter


def scan_csv(filename, niter = 10):
  start = time.time()
  fp = open(filename, 'r')
  for it in xrange(niter):
    fp.seek(0)
    for line in fp:
      row = tuple(float(f) for f in line.split(','))
      #print row
  fp.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter  

def scan_csv_numpy(filename, niter = 10):
  start = time.time()
  for it in xrange(niter):
    data = loadtxt(filename, dtype=float, delimiter=',')
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter  


def scan_pickle(filename, niter = 10):
  start = time.time()
  fp = open(filename, 'rb')
  for it in xrange(niter):
    fp.seek(0)
    nrow = pickle.load(fp)  # Read dataset length
    for i in xrange(nrow):
      row = pickle.load(fp)
      #print row
  fp.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter

def scan_zodb(filename, niter = 10):
  start = time.time()
  for it in xrange(niter):
    conn = ZODB.connection(filename)  # If this is outside loops: zodb caches objects..
    data = conn.root.data
    for row in data.values():  # TODO - check, is IOBTreeItems an iterator!
      # print row
      pass
    conn.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter


default_filetypes = (
  ('sql', mk_sqlite),
  ('h5', mk_table),
  ('csv', mk_csv),
  ('pickle', mk_pickle),
  ('opickle', mk_pickle_obj),
  ('zodb', mk_zodb),
)

default_scanners = (
  ('sqlite', 'sql', scan_sqlite),
  ('sqlite_obj', 'sql', scan_sqlite_obj),
  ('sqlalchemy_core', 'sql', scan_sqlalchemy_core),
  ('sqlalchemy_orm', 'sql', scan_sqlalchemy_orm),
  ('pytables', 'h5', scan_table),
  ('h5py', 'h5', scan_h5),
  ('csv', 'csv', scan_csv),
  ('csv_numpy', 'csv', scan_csv_numpy),
  ('cpickle', 'pickle', scan_pickle),
  ('cpickle_obj', 'opickle', scan_pickle),
  ('zodb', 'zodb', scan_zodb),
)

default_sizes = (1000, 5000, 10000, 50000, 100000, 500000, )

def mk_all_files (sizes, filetypes):
  for sz in sizes:
    for typ, mkf in filetypes:
      filename = 'data_%d.%s' % (sz, typ)
      print >>sys.stderr, "Writing %s" % filename
      mkf(filename, nrow=sz)
  print >>sys.stderr, "Done"


def scan_all (sizes, scanners, niter=10):
  nsizes = len(sizes)
  nscanners = len(scanners)
  df = pd.DataFrame(index=sizes, columns=(s[0] for s in scanners), dtype=float)
  for sz in sizes:
    for name, typ, scanf in scanners:
      filename = 'data_%d.%s' % (sz, typ)
      print >>sys.stderr, "Running %s on %s" % (name, filename)
      df.loc[sz,name] = scanf(filename, niter=niter)
  print >>sys.stderr, "Done"
  return df


def do_plots (meas, meas_cy = None, labels=('CPython',)):
  # Time plot
  figure()
  meas.plot(style='*-', legend=False)
  legend(meas.columns, fontsize='small')
  ylim(0,2)
  title('Load time vs size')
  xlabel('# records')
  ylabel('time (sec)')
  gcf().set_size_inches(7,4.75)
  savefig('wall-time.png', dpi=100, bbox_inches='tight')
  # Relative bars
  bars = meas.mean()/meas.mean().iloc[0]
  if meas_cy is not None:
    # meas_cy may be missing some columns, adjust
    avg_cy = meas_cy.mean()/meas.mean().iloc[0]
    bars_cy = pd.Series(index = bars.index)
    for c in avg_cy.index:
      bars_cy.loc[c] = avg_cy.loc[c]
  figure()
  ax = bars.plot(kind='barh', rot=45, width=0.4)
  if meas_cy is not None:
    ax_cy = bars_cy.plot(kind='barh', rot=45, width=0.4, color='g')
    for (c, dy) in zip(ax.containers, (-0.2, 0.2)):
      for r in c:
        r.set_y(r.get_y()+dy)
    draw()
  lg_h, lg_l = gca().get_legend_handles_labels()
  gca().legend(lg_h, labels, fontsize='small')
  draw()
  gca().set_yticklabels(bars.index, ha='right', va='top')
  yt = gca().get_yticks()
  print bars, bars_cy
  for i in range(len(bars)):
    if meas_cy is not None:
      text(bars.iloc[i]+2, yt[i]-0.2, '%.2f' % bars.iloc[i], ha='left', va='center', fontsize='small')
      if not isnan(bars_cy.iloc[i]):
        text(bars_cy.iloc[i]+2, yt[i]+0.2, '%.2f' % bars_cy.iloc[i], ha='left', va='center', fontsize='small')
    else:
      text(bars.iloc[i]+2, yt[i], '%.2f' % bars.iloc[i], ha='left', va='center')
  xlabel('Normalized time')
  title('Load times')
  gcf().set_size_inches(7,4)
  savefig('time-bar.png', dpi=100, bbox_inches='tight')


# if __name__ == '__main__':
#   mk_all_files(default_sizes, default_filetypes)
#   meas = scan_all(default_sizes, default_scanners)
#   do_plots(meas)
