from pylab import *
from numpy import *
import pandas as pd

import time

import cPickle as pickle
import tables as tbl
import sqlite3 as sql

from sqlalchemy import Table, Column, Integer
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


SqlBase = declarative_base()

class SqlRow(SqlBase):
  __tablename__ = 'data'
  c1 = Column(Integer, primary_key = True)  # Primary key is 'fake', but required by ORM
  c2 = Column(Integer)
  c3 = Column(Integer)
  c4 = Column(Integer)
  c5 = Column(Integer)


def mk_sqlite(filename, nrow = 10000):
  db = sql.connect(filename)
  c = db.cursor()
  c.execute("CREATE TABLE data (c1 REAL, c2 REAL, c3 REAL, c4 REAL, c5 REAL)")
  for i in xrange(nrow):
    c.execute("INSERT INTO data VALUES (?, ?, ?, ?, ?)", random.uniform(0.0, 100.0, 5))
  db.commit()
  db.close()


class TableRow(tbl.IsDescription):
  c1 = tbl.FloatCol()
  c2 = tbl.FloatCol()
  c3 = tbl.FloatCol()
  c4 = tbl.FloatCol()
  c5 = tbl.FloatCol()

def mk_table(filename, nrow = 10000):
  h5f = tbl.open_file(filename, mode='w', title='Benchmark')
  t = h5f.create_table('/', 'data', TableRow, 'Random data')
  row = t.row
  for i in xrange(nrow):
    for cname in ('c1', 'c2', 'c3', 'c4', 'c5'):
      row[cname] = random.uniform(0.0, 100.0)
    row.append()
  h5f.close()


def mk_csv(filename, nrow = 10000):
  with open(filename, 'w') as fp:
    for i in xrange(nrow):
      print >>fp, ','.join(str(ci) for ci in random.uniform(0.0, 100.0, 5))


def mk_pickle(filename, nrow = 10000):
  with open(filename, 'wb') as fp:
    pickle.dump(nrow, fp, pickle.HIGHEST_PROTOCOL)  # Write dataset length
    for i in xrange(nrow):
      row = tuple(random.uniform(0.0, 100.0, 5))
      pickle.dump(row, fp, pickle.HIGHEST_PROTOCOL)


class Row(object):
  def __init__(self, c1, c2, c3, c4, c5):
    self.c1 = c1
    self.c2 = c2
    self.c3 = c3
    self.c4 = c4
    self.c5 = c5

def mk_pickle_obj(filename, nrow = 10000):
  with open(filename, 'wb') as fp:
    pickle.dump(nrow, fp, pickle.HIGHEST_PROTOCOL)  # Write dataset length
    for i in xrange(nrow):
      row = Row(*tuple(random.uniform(0.0, 100.0, 5)))
      pickle.dump(row, fp, pickle.HIGHEST_PROTOCOL)


def scan_sqlite(filename, niter = 100):
  start = time.time()
  db = sql.connect(filename)
  for it in xrange(niter):
    c = db.cursor()
    c.execute("SELECT * FROM data")
    row = c.fetchone()
    while row is not None:
      #print row
      row = c.fetchone()
  db.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)    
  return t/niter

def scan_sqlite_obj(filename, niter = 100):
  start = time.time()
  db = sql.connect(filename)
  for it in xrange(niter):
    c = db.cursor()
    c.execute("SELECT * FROM data")
    row = c.fetchone()
    while row is not None:
      row_obj = Row(*row)
      #print row_obj
      row = c.fetchone()
  db.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)    
  return t/niter


def scan_sqlalchemy_orm(filename, niter = 100):
  start = time.time()
  engine = create_engine('sqlite:///%s' % filename)
  Session = sessionmaker(bind=engine)
  session = Session()
  for it in xrange(niter):
    for row in session.query(SqlRow):
      #print row
      pass
  session.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)    
  return t/niter

def scan_sqlalchemy_core(filename, niter = 100):
  start = time.time()
  engine = create_engine('sqlite:///%s' % filename)
  conn = engine.connect()
  for it in xrange(niter):
    for row in conn.execute(SqlRow.__table__.select()):
      #print row
      pass
  conn.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter


def scan_table(filename, niter = 10):
  start = time.time()
  h5f = tbl.open_file(filename, mode='r')
  for it in xrange(niter):
    nrows = len(h5f.root.data)
    data = h5f.root.data
    #for i in xrange(nrows):
    #  row = data[i]
    #  #print row
    for row in data:
      rowdata = row.fetch_all_fields()
      #print rowdata
  h5f.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter

def scan_csv(filename, niter = 10):
  start = time.time()
  fp = open(filename, 'r')
  for it in xrange(niter):
    fp.seek(0)
    for line in fp:
      row = tuple(float(f) for f in line.split(','))
      #print row
  fp.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter  

def scan_csv_numpy(filename, niter = 10):
  start = time.time()
  for it in xrange(niter):
    data = loadtxt(filename, dtype=float, delimiter=',')
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter  


def scan_pickle(filename, niter = 10):
  start = time.time()
  fp = open(filename, 'rb')
  for it in xrange(niter):
    fp.seek(0)
    nrow = pickle.load(fp)  # Read dataset length
    for i in xrange(nrow):
      row = pickle.load(fp)
      #print row
  fp.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter


default_filetypes = (
  ('sql', mk_sqlite),
  ('h5', mk_table),
  ('csv', mk_csv),
  ('pickle', mk_pickle),
  ('opickle', mk_pickle_obj),
)

default_scanners = (
  ('sqlite', 'sql', scan_sqlite),
  #('sqlite_obj', 'sql', scan_sqlite_obj),
  ('sqlalchemy_core', 'sql', scan_sqlalchemy_core),
  ('sqlalchemy_orm', 'sql', scan_sqlalchemy_orm),
  ('pytables', 'h5', scan_table),
  ('csv', 'csv', scan_csv),
  ('csv_numpy', 'csv', scan_csv_numpy),
  ('cpickle', 'pickle', scan_pickle),
  #('cpickle_obj', 'opickle', scan_pickle),
)

default_sizes = (1000, 5000, 10000, 50000, 100000, 500000, )

def mk_all_files (sizes, filetypes):
  for sz in sizes:
    for typ, mkf in filetypes:
      filename = 'data_%d.%s' % (sz, typ)
      print >>sys.stderr, "Writing %s" % filename
      mkf(filename, nrow=sz)
  print >>sys.stderr, "Done"


def scan_all (sizes, scanners, niter=10):
  nsizes = len(sizes)
  nscanners = len(scanners)
  df = pd.DataFrame(index=sizes, columns=(s[0] for s in scanners), dtype=float)
  for sz in sizes:
    for name, typ, scanf in scanners:
      filename = 'data_%d.%s' % (sz, typ)
      print >>sys.stderr, "Running %s on %s" % (name, filename)
      df.loc[sz,name] = scanf(filename, niter=niter)
  print >>sys.stderr, "Done"
  return df


def do_plots (meas):
  # Time plot
  figure()
  meas.plot(style='*-', legend=False)
  legend(meas.columns, fontsize='small')
  ylim(0,2)
  title('Load time vs size')
  xlabel('# records')
  ylabel('time (sec)')
  gcf().set_size_inches(6,4)
  savefig('wall-time.png', dpi=100, bbox_inches='tight')
  # Relative bars
  bars = meas.mean()/meas.mean().iloc[0]
  figure()
  bars.plot(kind='barh', rot=45)
  gca().set_yticklabels(bars.index, ha='right', va='top')
  yt = gca().get_yticks()
  for i in range(len(bars)):
    text(bars.iloc[i]+2, yt[i], '%.2f' % bars.iloc[i], ha='left', va='center')
  xlabel('Normalized time')
  title('Load times')
  gcf().set_size_inches(6,2.5)
  savefig('time-bar.png', dpi=100, bbox_inches='tight')


# if __name__ == '__main__':
#   mk_all_files(default_sizes, default_filetypes)
#   meas = scan_all(default_size, default_scanners)
#   do_plots(meas)
