#from pylab import *
#from numpy import *
#import pandas as pd

import sys

import time

import cPickle as pickle
import sqlite3 as sql
import persistent, transaction, ZODB

from sqlalchemy import Table, Column, Integer
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker


SqlBase = declarative_base()

class SqlRow(SqlBase):
  __tablename__ = 'data'
  c1 = Column(Integer, primary_key = True)  # Primary key is 'fake', but required by ORM
  c2 = Column(Integer)
  c3 = Column(Integer)
  c4 = Column(Integer)
  c5 = Column(Integer)


class Row(object):
  def __init__(self, c1, c2, c3, c4, c5):
    self.c1 = c1
    self.c2 = c2
    self.c3 = c3
    self.c4 = c4
    self.c5 = c5

# In separate file, so it works across bench and bench_pypy
from zrow import ZRow


def scan_sqlite(filename, niter = 100):
  start = time.time()
  db = sql.connect(filename)
  for it in xrange(niter):
    c = db.cursor()
    c.execute("SELECT * FROM data")
    row = c.fetchone()
    while row is not None:
      #print row
      row = c.fetchone()
  db.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)    
  return t/niter

def scan_sqlite_obj(filename, niter = 100):
  start = time.time()
  db = sql.connect(filename)
  for it in xrange(niter):
    c = db.cursor()
    c.execute("SELECT * FROM data")
    row = c.fetchone()
    while row is not None:
      row_obj = Row(*row)
      #print row_obj
      row = c.fetchone()
  db.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)    
  return t/niter


def scan_sqlalchemy_orm(filename, niter = 100):
  start = time.time()
  engine = create_engine('sqlite:///%s' % filename)
  Session = sessionmaker(bind=engine)
  session = Session()
  for it in xrange(niter):
    for row in session.query(SqlRow):
      #print row
      pass
  session.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)    
  return t/niter

def scan_sqlalchemy_core(filename, niter = 100):
  start = time.time()
  engine = create_engine('sqlite:///%s' % filename)
  conn = engine.connect()
  for it in xrange(niter):
    for row in conn.execute(SqlRow.__table__.select()):
      #print row
      pass
  conn.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter


def scan_csv(filename, niter = 10):
  start = time.time()
  fp = open(filename, 'r')
  for it in xrange(niter):
    fp.seek(0)
    for line in fp:
      row = tuple(float(f) for f in line.split(','))
      #print row
  fp.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter  

# def scan_csv_numpy(filename, niter = 10):
#   start = time.time()
#   for it in xrange(niter):
#     data = loadtxt(filename, dtype=float, delimiter=',')
#   t = time.time() - start
#   print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
#   return t/niter  


def scan_pickle(filename, niter = 10):
  start = time.time()
  fp = open(filename, 'rb')
  for it in xrange(niter):
    fp.seek(0)
    nrow = pickle.load(fp)  # Read dataset length
    for i in xrange(nrow):
      row = pickle.load(fp)
      #print row
  fp.close()
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter

def scan_zodb(filename, niter = 10):
  start = time.time()
  for it in xrange(niter):
    conn = ZODB.connection(filename)  # If this is outside loops: zodb caches objects..
    data = conn.root.data
    for row in data.values():  # TODO - check, is IOBTreeItems an iterator!
      # print row
      pass
    transaction.abort()   # Pypy needs this, see next line
    conn.close()  # Fails in Pypy with ConnectionStateError: Cannot close a connection joined to a transaction
  t = time.time() - start
  print >>sys.stderr, "%.2f total, %.3f per iter" % (t, t/niter)
  return t/niter


default_scanners = (
  ('sqlite', 'sql', scan_sqlite),
  ('sqlite_obj', 'sql', scan_sqlite_obj),
  ('sqlalchemy_core', 'sql', scan_sqlalchemy_core),
  ('sqlalchemy_orm', 'sql', scan_sqlalchemy_orm),
  ('csv', 'csv', scan_csv),
  #('csv_numpy', 'csv', scan_csv_numpy),
  ('cpickle', 'pickle', scan_pickle),
  ('cpickle_obj', 'opickle', scan_pickle),
  ('zodb', 'zodb', scan_zodb),
)

default_sizes = (1000, 5000, 10000, 50000, 100000, 500000, )


def scan_all (sizes, scanners, niter=10):
  nsizes = len(sizes)
  nscanners = len(scanners)
  #df = pd.DataFrame(index=sizes, columns=(s[0] for s in scanners), dtype=float)
  df = {}
  for sz in sizes:
    for name, typ, scanf in scanners:
      filename = 'data_%d.%s' % (sz, typ)
      print >>sys.stderr, "Running %s on %s" % (name, filename)
      df[(sz,name)] = scanf(filename, niter=niter)
  print >>sys.stderr, "Done"
  return df


if __name__ == '__main__':
  meas = scan_all(default_sizes, default_scanners)
  print meas

